const { ccclass } = cc._decorator;

@ccclass
export default class Utils {
    public static deepCopy(obj): any {
        let _obj = Array.isArray(obj) ? [] : {}
        for (let i in obj) {
            _obj[i] = typeof obj[i] === 'object' ? this.deepCopy(obj[i]) : obj[i]
        }
        return _obj
    }

    public static log(param: { [k: string]: any } | string, prefix: string = "Blade:", suffix: string = "", splitChar: string = ',') {
        let str = prefix;
        if (typeof param == 'string') {
            str += param;
        }
        else {
            let count = 0;
            for (let key in param) {
                let v = param[key];
                str += count > 0 ? splitChar : "";
                if (typeof v == 'object') {
                    str += `${key}:${JSON.stringify(v)}`;
                }
                else {
                    str += `${key}:${v}`;
                }
                count++;
            }
        }
        str += suffix;
        if (CC_EDITOR) {
            Editor.log(str);
        }
        else {
            cc.log(str);
        }
    }

    public static logError(param: { [k: string]: any } | string, prefix: string = "Blade:", suffix: string = "", splitChar: string = ',') {
        let str = prefix;
        if (typeof param == 'string') {
            str += param;
        }
        else {
            let count = 0;
            for (let key in param) {
                let v = param[key];
                str += count > 0 ? splitChar : "";
                if (typeof v == 'object') {
                    str += `${key}:${JSON.stringify(v)}`;
                }
                else {
                    str += `${key}:${v}`;
                }
                count++;
            }
        }
        str += suffix;
        if (CC_EDITOR) {
            Editor.warn(str);
        }
        else {
            console.error(str);
        }
    }

    /**
     * 返回 [min,max)之间的随机浮点数
     * @param min 
     * @param max 
     * @returns 
     */
    public static randomFloat(min: number | Array<number>, max?: number): number {
        if (Array.isArray(min)) {
            max = Math.max(...min);
            min = Math.min(...min);
        }
        return Math.random() * (max - min) + min;
    }

    /**
     * 返回 [min,max]之间的随机整数
     * @param min 
     * @param max 
     * @returns 
     */
    public static randomInt(min: number | Array<number>, max?: number): number {
        if (Array.isArray(min)) {
            max = Math.max(...min);
            min = Math.min(...min);
        }
        return Math.floor(Math.random() * (max - min + 1)) + min;
    }

    public static getValueByWeight(arrValue: Array<number>, arrWeight: Array<number>) {
        let total = arrWeight.reduce((total, v) => {
            return total + v;
        })
        let random = this.randomFloat(0, total);
        // cc.log("getValueByWeight  random range===>>> ", random)
        let temp = 0;
        for (let i = 0; i < arrValue.length; i++) {
            const weight = arrWeight[i] + temp;
            if (random <= weight) {
                // cc.log("getValueByWeight  随到了===>>> ", arrValue[i])
                return arrValue[i];
            }
            temp = weight;
        }
        // cc.log("getValueByWeight 没随到 ")
        return null;
    }

    public static setNodesVisible(nodes: cc.Node[], visible: boolean) {
        if (Array.isArray(nodes) && nodes.length > 0) {
            nodes.forEach(n => {
                if (n.isValid) {
                    n.active = visible;
                }
            })
        }
        else {
            cc.error(`Blade:参数错误！`)
        }
    }
    public static isInPolygon(checkPoint: cc.Vec2, polygonPoints: cc.Vec2[]) {
        var counter = 0;
        var i: number;
        var xinters;
        var p1: cc.Vec2, p2: cc.Vec2;
        var pointCount = polygonPoints.length;
        p1 = polygonPoints[0];

        for (i = 1; i <= pointCount; i++) {
            p2 = polygonPoints[i % pointCount];
            if (
                checkPoint.x > Math.min(p1.x, p2.x) &&
                checkPoint.x <= Math.max(p1.x, p2.x)
            ) {
                if (checkPoint.y <= Math.max(p1.y, p2.y)) {
                    if (p1.x != p2.x) {
                        xinters = (checkPoint.x - p1.x) * (p2.y - p1.y) / (p2.x - p1.x) + p1.y;
                        if (p1.y == p2.y || checkPoint.y <= xinters) {
                            counter++;
                        }
                    }
                }
            }
            p1 = p2;
        }
        if (counter % 2 == 0) {
            return false;
        } else {
            return true;
        }
    }
    // 判断一个点是否在三角形内
    public static isInTriangle(point: cc.Vec2, triA: cc.Vec2, triB: cc.Vec2, triC: cc.Vec2) {
        let AB = triB.sub(triA), AC = triC.sub(triA), BC = triC.sub(triB), AD = point.sub(triA), BD = point.sub(triB);
        //@ts-ignore
        return (AB.cross(AC) >= 0 ^ AB.cross(AD) < 0) && (AB.cross(AC) >= 0 ^ AC.cross(AD) >= 0) && (BC.cross(AB) > 0 ^ BC.cross(BD) >= 0);
    }

    // 多边形 三角切割
    public static splitePolygon(points: cc.Vec2[]): number[] {
        if (points.length <= 3) return [0, 1, 2];
        let pointMap: { [key: string]: number } = {};     // point与idx的映射
        for (let i = 0; i < points.length; i++) {
            let p = points[i];
            pointMap[`${p.x}-${p.y}`] = i;
        }
        const getIdx = (p: cc.Vec2) => {
            return pointMap[`${p.x}-${p.y}`]
        }
        points = points.concat([]);
        let idxs: number[] = [];

        let index = 0;
        while (points.length > 3) {
            let p1 = points[(index) % points.length]
                , p2 = points[(index + 1) % points.length]
                , p3 = points[(index + 2) % points.length];
            let splitPoint = (index + 1) % points.length;

            let v1 = p2.sub(p1);
            let v2 = p3.sub(p2);
            if (v1.cross(v2) < 0) {      // 是一个凹角, 寻找下一个
                index = (index + 1) % points.length;
                continue;
            }
            let hasPoint = false;
            for (const p of points) {
                if (p != p1 && p != p2 && p != p3 && this.isInTriangle(p, p1, p2, p3)) {
                    hasPoint = true;
                    break;
                }
            }
            if (hasPoint) {      // 当前三角形包含其他点, 寻找下一个
                index = (index + 1) % points.length;
                continue;
            }
            // 找到了耳朵, 切掉
            idxs.push(getIdx(p1), getIdx(p2), getIdx(p3));
            points.splice(splitPoint, 1);
        }
        for (const p of points) {
            idxs.push(getIdx(p));
        }
        return idxs;
    }

    /** 计算uv, 锚点都是中心 */
    public static computeUv(points: cc.Vec2[], width: number, height: number) {
        let uvs: cc.Vec2[] = [];
        for (const p of points) {
            // uv原点是左上角
            let x = this.clamp(0, 1, (p.x + width / 2) / width);
            let y = this.clamp(0, 1, 1. - (p.y + height / 2) / height);
            uvs.push(cc.v2(x, y));
        }
        return uvs;
    }

    public static clamp(a: number, b: number, x: number) {
        if (a > b) {
            let t = a;
            a = b;
            b = t;
        }
        if (x < a) return a;
        if (x > b) return b;
        return x;
    }
    /**
     * @是否是长屏幕
     * @returns 
     */
    public static isLongScreen() {
        if (cc.winSize.height / cc.winSize.width < 1.9) {
            return false;
        }
        return true;
    }
    public static stringFormat(...args): string {
        if (args.length == 0) {
            return "";
        }
        let str: string = args[0];
        for (let i = 1; i < args.length; i++) {
            str = str.replace(new RegExp("\\{" + (i - 1) + "\\}", "g"), args[i]);
        }

        return str;
    }


}
